# TODO: Fill out this file with information about your package

# HINT: Go back to the object-oriented programming lesson "Putting Code on PyPi" and "Exercise: Upload to PyPi"

# HINT: Here is an example of a setup.py file
# https://packaging.python.org/tutorials/packaging-projects/

from setuptools import setup

setup(name='jzrki_linear_system',
      version='0.1',
      description='Solve a System of N-dimensional Linear Equations',
      packages=['jzrki_linear_system'],
      author = 'James L Zhang',
      author_email = 'jzr.kingsta@outlook.com',
      zip_safe=False)
